/*! \file test.c

	\brief Yksinkertainen demo FreeRTOS-reaaliaikakyttjrjestelmlle

	http://www.freertos.org

	Demo lukee raudan nppimi sek kyttjrjestelmn kellolaskuria ja pivitt nit LCD-nytlle ja ledeille.
	Kutakin kohdetta hoitaa oma taskinsa muista riippumatta. Kommunikointi tskien vlill on toteutettu
	kahta viestijonoa kytten. Demo on yksinkertainen ja hydynt vain pienen murto-osan kyttjrjestelmn
	ominaisuuksista. Demoa ei myskn ole edes pyritty optimoimaan joten sen resurssinkulutusta voi varmaan
	pienent viel huomattavasti. Tm sovellus kuitenkin hyvin mahtuu kytettyyn ympristn.

	Demo on buildattu kytten AVR Studio v 4.18 build 684 ja Atmel STK500-testilauta.
	Kohdeprosessori on ATMega32, mik on huomioitava kokeiltaessa muilla kohteilla.
	Porttaus pitisi toimia kaikissa Mega-sarjan prosessoreissa joissa on samanlainen
	timerisysteemi kuin Mega32:ssa (16-bittinen Timer/counter 1 jossa compare output).
	Kaikki kyttiksen porttaus on tehty tiedostoon port.c ja projektissa kytetn
	RTOS:in mukana tulevaa heap-allokaattoria nro 1 (lue tarkemmin RTOS-dokumentaatiosta)
	Tt demoa varten FreeRTOS:in oletushakemistorakenne on litistetty yhteen tasoon 
	ja porttaukset sek demot kaikille muille prosessoreille on poistettu tilan sstmiseksi

	Demo on toteutettu Atmelin STK500-kehitysalustalle, mutta sen pitisi olla helposti
	muokattavissa toimimaan miss vaan sill laitevaatimukset ovat vaatimattomat.
	Portti A on varattu LCD-nytn ohjaukseen. Signaalilinjat ja niiden kytt selvivt
	lcd-nytn ohjauskirjastosta (tiedostot lcd.h ja lcd.c)
	Portti B on varattu led-nytille (8 kpl). STK 500 ledit palavat kun ohjaus on 0
	Portista D luetaan kortin painikkeita (8 kpl). Painike on aktiivinen 0-suuntaan, eli kun yhtn
	nppint ei ole painettu, portista luetaan 0xff.

	Tm demo on runsaasti kommentoitu sit silmllpiten, ett aiheesta kiinnostuneet
	mutta C-kielen suhteen alussa olevatkin psevt kiinni tapahtumen kulkuun.

	Vastailen mielellni tt koskeviin kysymyksiin cnc-foorumilla. --> Kremmen <--

*/

// WinAVR-kntjn vakiokirjaston AVR-libc mukaan otettavat modulit

#include <avr\io.h>
#include <stdlib.h>


// FreeRTOS-kyttjrjestelmn headerit
// Sovelluksen ei koskaan tarvitse ottaa mukaan enemp kuin nm.
// Huomaa, ett "semphr.h" on kommentoitu pois, koska tm demo ei
// tss muodossa kyt semaforeja eik mutexeja

#include "FreeRTOS.h"
#include "task.h"
// #include "semphr.h"
#include "queue.h"


// Peter Fleuryn LCD-paketti HD44780-yhteensopivalle nytlle
// Koodia on sdetty toimimaan 4*16 merkin Displaytech 164-nytn kanssa 4-bitin moodissa

#include "lcd.h"

// LCD-taskin viestijonon tarvitsemat esittelyt

#define QLEN 5				// max viiden viestin jono

typedef struct message {	// jonoon laitetaan tllaisia viestej
	uint8_t	sender;
	uint32_t payload;
	} message_t, *message_p;

#define SENDER_INPUT_TASK	1	// LCD-taskille vlitetn tieto siit, kuka on lhettnyt viestin
#define SENDER_TICK_COUNTER	2	// Nin LCD-taski tiet mit viestin data on ja miten se nytetn

// Inputtaski lhett havaitsemiaan nppimen painalluksia sek LCD-taskille ett led-taskille
// Tst syyst sille pit taskin luonnissa vlitt kahden jonon osoitteet.
// luontiparametri on tyyppi (void *) josa voidaan vlitt ainoastaan 1 osoite. Tytyy siis
// esitell struktuurityyppi johon voidaan sijoittaa kahden jonon osoitteet ja vlitt sen osoite
// kynnistysparametrina. Vlittmll taskin kynnistysparametrina struktuurin osoite, voidaan 
// struktuuri mritell sisltmn mielivaltaisen paljon alustusdataa.

typedef struct doubleque {	// kahden jonon osoitteet sisltv struktuurityyppi (huom vain tyyppi, ei viel muuttuja!)
	xQueueHandle que1;
	xQueueHandle que2;
	} doubleque_t, *doubleque_p;	//esitelln samalla vaivalla lisksi pointterityyppi vaikkei tss tarvitalaan.


/*! =======================led-nytn ohjaustaski ===================================*/
//	Lukee jonosta nyttkskyja ja ohjaa sen perusteella led-nytt
//	nppinportin lukutaski lhett nppintietoja jonoon josta ne luetaan ja aletaan vilkuttaa vastaavaa ledi
//	kunnes jonosta saadaan uusi nppin.

void LedTask (void * pvParameters) {	// RTOS mrittelee taskit aina tmn nkisiksi funktioiksi

	xQueueHandle command_queue;	// saapuvien viestien jonon "kahva" eli osoitin jonoon. RTOS:in mrittelem tyyppi
	uint8_t	leds, sw;			// ledit ja vilkutuksen apumuuttuja

	// kynnistysparametrina saatiin jono johon nytettv tavara tulee
	// koska jonon kahva on tyyppi xQueueHandle ja kutsuparametri pvParameters on tyyppi void pointer (void *)
	// suoritetaan eksplisiittinen typecast eli komennetaan sijoituslauseessa parametri olemaan
	// haluttua tyyppi. Se tapahtuu nin:

	command_queue = (xQueueHandle) pvParameters;	// (tss toimii ilman typecastiakin, mutta ei en alempana InputTaskissa)

	// Alustetaan lediportti. Demossa ei vaivauduta mrittelemn porttia tmn kummemmin vaan
	// kovakoodataan se kylmsti thn suoraan. Tuotantokoodissa portti pitisi mritt #define kytten
	// jossain yleisess headerissa jotta se on helppo vaihtaa tarvittaessa. Nyt antaa olla nin.

	PORTB = 0xff;
	DDRB  = 0xff;

	// taskin muuttujille alkuarvot. Lokaalit muuttujat tallennetaan pinomuistissa, eik kntj voi antaa niille
	// taattuja alkuarvoja. Sen takia pit tehd itse jos sill on vli. Mikli nm olisi mritelty static-
	// muuttujiksi, ne olisivat menneet datasegmenttiin joka nollataan kynnistyksess.

	leds = 0xff;	// ledit alussa pimeiksi
	sw = 0;			// vilkutusswitchi alkutilaan 

	// taski pyrii tss silmukassa maailmanloppuun asti koskaan palaamatta
	for ( ; ; ) {

		// luetaan jonosta command_queue viesti jos siell on ja laitetaan sen sislt muuttujaan leds.
		// Jos jonossa ei ole juuri nyt yhtn viesti, ei jd odottelemaan (viimeinen parametri on 0).
		if ( xQueueReceive( command_queue, &leds, 0 )) {

			PORTB = leds;	// jos viesti saatiin (xQueueReceive palautti != 0) niin pivitetn portti heti uudella arvolla
			sw = 0;			// ja tahdistetaan vilkutus pivityksen kanssa
		}

		// odotetaan 5 kyttjrjestelmn aikaskedulerin kierrosta. Tarkka aika riippuu muuttujan configTICK_RATE_HZ
		// arvosta tiedostossa FreeRTOSConfig.h Viive on 5 kellon "tikityst".
		// vTaskDelay EI varaa prosessoria vaan laittaa tmn taskin nukkumaan kunnes viidennen kellokeskeytyksen jlkeen
		// taski taas hertetn uniltaan. Sin aikana prosessori on kaikkien muiden taskien kytss.
		vTaskDelay(5);

		// kytkimen sw tilasta riippuen nytetn kytkint tai pime jolloin kytkint vastaava ledi vilkkuu
		if ( sw ) {
			PORTB = leds;
		} else {
			PORTB = 0xff;
		}
		sw = ~sw;	// knnetn kytkint
	}
}

/*! =======================LCD-nytn ohjaustaski ===================================*/
//	Lukee jonosta nyttkskyja ja ohjaa sen perusteella LCD-nytt
//	Kaikki perusasiat ovat juuri samoin kuin leditaskissa ylempn

void LCDTask (void * pvParameters) {

	char buffer[10];	// nytn merkkipuskuri. Varattu 10 tavua [0...9]. Max 9 kirjainta + terminoiva NULL

	xQueueHandle command_queue;	// jonon kahva
	message_t msg;				// saapuvien viestien tallennuspaikka. Ylempn esitelty message_t -tyypinen struktuuri
	
	command_queue = (xQueueHandle) pvParameters;	// kynnistysparametrina saatiin jono johon nytettv tavara tulee

	// Alustetaan LCD-nytt ja heitetn joku splash screeni
	// Nm kaikki funktiot lytyvt lcd-kirjastosta (lcd.h ja lcd.c)
	lcd_init(LCD_DISP_ON_CURSOR);
	lcd_clrscr();
	lcd_gotoxy(0,0);
	lcd_puts_P("Pieni RTOS demo");	// Nm tekstit tulevat prosessorin KOODImuistista eivtk sy rammia!
	lcd_gotoxy(0,1);
	lcd_puts_P("RTOS v 6.0.1");
	lcd_gotoxy(0,2);
	lcd_puts_P("(c) 2010 Kremmen");




	// taas ikuinen silmukka
	for ( ; ; ) {

		// luetan viesti jonosta. Tll kerralla odotellaan tss kunnes sadaan viesti. Kun ei ole muutakaan tekemist.
		// Viiveen arvo portMAX_DELAY pist odottamaan ikuisesti
		if ( xQueueReceive( command_queue, &msg, portMAX_DELAY )) {	

			// message-struktuurin payload-muuttujassa on nytettv tavara.
			// Tss tapauksessa joko kellotikitysten mr tai nppin. Tulkitaan molemmat numeroksi ja
			// muutetaan numero printattavaksi merkkijonoksi tekstijonomuuttujaan buffer
			itoa(msg.payload, buffer, 10);	

			switch (msg.sender) {				// tsekataan kenelt viesti tuli ja kirjoitellaan nytlle sen mukaan
				case SENDER_INPUT_TASK: {		// lhettj oli nppimisttaski
					lcd_gotoxy(0,2);
					lcd_puts_P("Got             ");
					lcd_gotoxy(4,2);
					lcd_puts(buffer);
					break;						// muistetaan breikata, ettei valuta lpi seuravaan caseen!
				}
				case SENDER_TICK_COUNTER: {		// lhettj oli kellotaski
					lcd_gotoxy(0,3);
					lcd_puts_P("Cnt             ");
					lcd_gotoxy(4,3);
					lcd_puts(buffer);
				}
			}
		} else {	// thn ei ikin pitisi pty tai 'Houston meill on ongelma...'
			lcd_clrscr();
			lcd_puts_P("JONO SRKYI!");
		}		
	}
}// end task LCDTask


/*! ======================= Painonppinten lukutaski ===================================*/
//	Lukee STK500:n (tai mink vaan) painikkeita portista D ja viestitt ne LCD-taskille ja ledeille

// Tm taski saa kynnistysparametrina osoittimen struktuuriin josta pit purkaa kahden jonon kahvat
// oikeille paikoilleen.

void InputTask ( void *pvParameters) {

	xQueueHandle	LCD_queue, led_queue;	// lhtevien jonojen kahvat
	message_t		msg;					// viestistruktuuri LCD-nytlle lhtevi viestej varten

    uint8_t			napit1, napit2;			// koskettimien luvussa kytetyt muuttujat

	// Alustetaan kovakoodattu lukuportti D
	// Alustus voi olla tarpeeton mutta varmistetaan kuitenkin

	PORTD = 0x00;
	DDRD  = 0x00;


	// Nyt tarkkana, tulee pakollinen typecast eik kaikkein helpointa lajia:
	// InputTask saa yhden void pointer (void *) tyyppisen parametrin kuten kaikki muutkin taskit.
	// LCD- ja led taskeille tuo pointteri sislsi suoraan jonon kahvan (joka sekin on oikeasti pointteri)
	// ja riitti kun sen pani muistiin jonon kahvamuuttujaan.
	// Tss parametrina tulee pointteri doubleque_t -tyyppiseen struktuuriin joka sislt kahden jonon kahvat.
	// Nuo kahvat on kopioitava ulos struktuurista omiin kahvamuuttujiinsa. Sit varten pit pysty osoittamaan
	// niit struktuurin jsenmuuttujia (que1 ja que2) joissa kahvat ovat. Thn tarvitaan eksplisiittinen typecast.
	// Kahta seuraavaa koodirivi kannattaa tutkia kunnes ymmrt mit niiss tehdn. Lhdemateriaalista lunttaaminen 
	// ei ole kielletty :). Niiss siis vitetn kntjlle, ett pvParameters osoittaa nyt doubleque_t -tyyppiseen
	// muuttujaan ja on OK sen kautta osoittaa struktuurin jsenmuuttujia que1 ja que2. Ilman typecastia kntj
	// ei tt usko.

	LCD_queue = ((doubleque_t*) pvParameters)->que1;	// kynnistysparametrina saatiin lcd-taskille menev jono
	led_queue = ((doubleque_t*) pvParameters)->que2;	// kynnistysparametrina saatiin led-taskille menev jono

	// taas ikuiseen silmukkaan
	for ( ; ; ) {

		// Yksinkertainen kosketinvrhtelyn poisto		
	    do {
    		napit1 = PIND;		// luetaan napit ja
    		vTaskDelay(1);		// odotetaan yksi tikitys
    		napit2 = PIND;      // luetaan napit uudelleen vertailumuuttujaan
		} while ((napit1 == 0xff) || (napit1 != napit2)); // jatketaan etenpin jos jotain on painettu ja napit ovat samat

		// odotetaan napin vapautusta jottei ohjelma rynt
		do {
			napit2 = PIND;
			vTaskDelay(1);
		} while (napit2 != 0xff);

		// nyt stabiilit koskettimet ovat muutujassa napit1;

		xQueueSendToBack(led_queue, &napit1, portMAX_DELAY);	// laitetaan viesti napeista menemn leditaskille pin.

		// LCD-taskille joudutaan rakentamaan viesti jossa tunnistetaan lhettj, ja painetun napin tieto nytettvksi

		msg.sender = SENDER_INPUT_TASK;		// viestin lhettjksi input-taski
		napit1 = ~napit1;					// knnetn napit1 ettei turhaan lhetet knteisarvoa.
		msg.payload = (uint32_t) napit1;	// varmistetaan typecastilla, ett tule etunollia (kntj kyll hoitaisi)
		
		// Laitetaan viesti menemn. Odotusaika opn pantu maksimiin, eli lhetys odottaa vuoroaan niin kauan,
		// ett jonossa on tilaa
		xQueueSendToBack(LCD_queue, &msg, portMAX_DELAY);
	}
} // end task InputTask

/*! ======================= Reaaliaikakellon lukutaski ===================================*/
//	Lukee RTOS-kyttiksen kellon tikityksi (tickej) ja raportoi ne nytttaskille esitettvksi

void TickCounterTask  ( void *pvParameters) {

	xQueueHandle	out_queue;	// samaa kamaa kuin muissakin taskeissa
	message_t		msg;		// ditto

	out_queue = (xQueueHandle) pvParameters;	// kynnistysparametrina saatiin lcd-taskille menev jono

	// looppi
	for ( ; ; ) {
		// odotellaan aina 10 tikityst kierrosten vlill, ettei lhetell hulluna laskurin lukemia
		vTaskDelay(10);

		msg.sender = SENDER_TICK_COUNTER;					// merkitn viestin lhettj tiedoksi
		msg.payload = (uint32_t) xTaskGetTickCount();		// viestin sisltn suoraan kutsulla tikityslaskurin arvo
		xQueueSendToBack(out_queue, &msg, portMAX_DELAY);	// ja viesti menemn
	}

}// end task TickCounterTask


/*! =======================          Demon pohjelma    ===================================*/
//	Pohjelmassa luodaan taskien vliseen viestintn tarkoitetut jonot (2 kpl)
//	sek demoon osallistuvat taskit, joiden koodi on siis ylempn tss tiedostossa.
//	Todellisessa sovelluksessa jokainen taski olisi melko varmasti omassa kooditiedostossaan, jonka headeri olisi
//	esitelty tmn tiedoston alussa. Havainnollisuuden vuoksi nin ei ole nyt tehty.

int main() {

	doubleque_t		dque;

	// Luodaan viestijonot LCD-taskille ja led-taskille menevi viestej varten
	// xQueueCreate-funktion paluuarvona saadaan kahva luotuun jonoon. Muistin sstmiseksi ei varata niille 
	// erillisi muuttujia vaa sijoitetaan ne suoraan dque-struktuurin muuttujiin que1 ja que2. Tm struktuuri
	// on tydennettv joka tapauksessa Input-taskin parametrinvlityst varten

	dque.que1 = xQueueCreate( QLEN, sizeof(message_t) );	// pitisi varmistaa, ettei paluuarvo == NULL mutta uskotaan
	dque.que2 = xQueueCreate( QLEN, sizeof(uint8_t) );

	// Luodaan seuraavaksi tyskentelijtaskit RTOS:in ajonaikaisiin tietorakenteisiin perkkisill xTaskCreate -kutsuilla.
	// Ensimminen parametri on osoitin taskina toimivaan funktioon.
	// Toinen parametri on taskin nimi kyttjrjestelmn sisisell listalla.
	// Kolmas parametrin on KRIITTISEN TRKE taskille varattavan pinomuistin koko.
	//  Kun tss taskit ovat kovin yksinkertaisia niille on varattu RTOS-systeemin minimipino.
	//  Pinon kasvaminen yli sille varatun mrn joko kaataa koko systeemin heti tai ainakin korruptoi sen pahasti. Sita
	//  siis yksinkertaisesti EI SAA TAPAHTUA! RTOSissa on funktiot taskin pinon koon mittaamiseen testien aikana.
	// Neljs parametri on osoitin taskifunktiolle vlitettvn kynnistysparametriin. Nihin laitetaan tss siis joko
	//  sopivan jonon kahva suoraan tai Input-taskin tapauksessa dque-struktuurin osoite jotta saadaan molemmat jonot 
	//  vlitetty sille.
	// Viides parametri on taskin suhteellinen prioriteetti. Isompi luku - isompi prioriteetti. RTOS takaa, ett isoin
	//  ajokelpoinen prioriteetti saa aina ajovuoron. Nm ovat demokraattisesti kaikki 1.
	// Kuudes ja viimeinen parametri on pointteri muuttujaan johon palautetaan luodun taskin kahva. Tm on tarpeellinen
	// tieto mikli taskiin aiotaan kohdistaa myhemmin operaatioita (esim Delete). Tss ei tehd muuttujalla mitn
	// joten vlitetn parametriksi NULL. Luonti on tarpeeksi fiksu ymmrtkseen, ettei tarvitse palauttaa nyt mitn.

	xTaskCreate( LedTask, ( signed char * ) "LED", configMINIMAL_STACK_SIZE, dque.que2, 1, NULL );
	xTaskCreate( LCDTask, ( signed char * ) "LCD", configMINIMAL_STACK_SIZE, dque.que1, 1, NULL );
	xTaskCreate( InputTask, ( signed char * ) "INP", configMINIMAL_STACK_SIZE, &dque, 1, NULL );
	xTaskCreate( TickCounterTask, ( signed char * ) "TCK", configMINIMAL_STACK_SIZE, dque.que1, 1, NULL );

	// Kaikki tarvittava on nyt rakennettu pystyyn ja voidaan aloittaa.

	// Seuraava kutsu kynnist reaaliaikaskedulerin.
	// Suoritus sukeltaa funktioon kuin Liisa kaninkoloon eik koskaan palaa. Sen sijaan skeduleri alkaa jakamaan
	// prosessorin suoritusaikaa luotujen taskien vlill kuin croupier kortteja kasinon pokeripydss.

	vTaskStartScheduler();
}
